#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MAIN 
#include <boost/test/unit_test.hpp>
#include <sdsl/balanced_parentheses_support_j.hpp>
#include <sdsl/testutils.hpp>
#include <stack>
#include <vector>

typedef sdsl::bit_vector::size_type size_type;

using namespace boost::unit_test;
using namespace boost::test_tools;
using namespace sdsl;

BOOST_AUTO_TEST_SUITE( balanced_parentheses_support_j_performance_test_suite )

void create_random_tree(size_t begin, size_t end, size_t max_degree, bit_vector &v){
	if(begin>=end)
		return;
//	BOOST_MESSAGE("begin = "<<begin<<" end = "<<end);
	size_t interval_size = end-begin;
	assert(interval_size%2==0 and interval_size>1);
	size_t degree = ((rand())%max_degree)+1;
	if(degree*2 > interval_size)
		degree = interval_size/2;
	for(size_t i=0; i<degree-1; ++i){
		interval_size = end-begin;
		size_t size = rand()%(interval_size+1-2*(degree-i-1));
		if(size==0)
			size=2;
		if(size%2==1)
			++size;
		v[begin] = 1;
		v[begin+size-1] = 0;
		create_random_tree(begin+1, begin+size-1, max_degree, v);
		begin = begin+size;
	}
	v[begin] = 1;
	v[end-1] = 0;
	create_random_tree(begin+1, end-1, max_degree, v);
}

BOOST_AUTO_TEST_CASE( correctness_test ){
	unit_test_log.set_threshold_level( log_messages );
	BOOST_MESSAGE( "Running performance test for balanced_parentheses_support_j..." );
	const size_type test_cases = 4;
	bit_vector v[test_cases];

	v[0].resize(10000000);
	create_random_tree(0, v[0].size(), 4, v[0]);
	BOOST_MESSAGE("test case 1: random tree of size "<<v[0].size()<<" and max branching factor "<<4<<" was created");
	v[1].resize(10000000);
	create_random_tree(0, v[1].size(), 1, v[1]);
	BOOST_MESSAGE("test case 2: random tree of size "<<v[1].size()<<" and max branching factor "<<1<<" was created");
	v[2].resize(10000000);
	create_random_tree(0, v[2].size(), 26, v[2]);
	BOOST_MESSAGE("test case 3: random tree of size "<<v[2].size()<<" and max branching factor "<<26<<" was created");
	v[3].resize(10000000);
	create_random_tree(0, v[3].size(), 128, v[3]);
	BOOST_MESSAGE("test case 4: random tree of size "<<v[3].size()<<" and max branching factor "<<128<<" was created");


	size_type k=0;

	for(size_type tc=0; tc<test_cases; ++tc){
		BOOST_MESSAGE("test case "<<tc+1);
		std::stack<size_type> s;
		std::vector<size_type> matches(v[tc].size(),0);
		std::vector<size_type> excess(v[tc].size(),0);
		const size_type size = v[tc].size();
		// calculate find_close answers
		for(size_type i=0; i < size; ++i){
			if(v[tc][i])
				s.push(i);
			else{
				BOOST_REQUIRE(!s.empty());
				matches[s.top()] = i;
				s.pop();
			}
		}
		BOOST_REQUIRE(s.empty());
		// calculate find_open answers
		for(size_type i=0; i < size; ++i){
			if(!v[tc][size-i-1])
				s.push(size-i-1);
			else{
				BOOST_REQUIRE(!s.empty());
				matches[s.top()] = i;
				s.pop();
			}
		}
		BOOST_REQUIRE(s.empty());

		// create balanced_parentheses_support_j for v
		balanced_parentheses_support_j<> bp_support(&v[tc]);
		stop_watch stopwatch;
		// meassure excess 
		stopwatch.start();
		for(size_type i=0; i < size; ++i){
			if(v[tc][i])
				k += bp_support.excess(i);
			else
				k -= 1;
		}
		stopwatch.stop();
		BOOST_MESSAGE("Time for "<<size<<" excess calls "<< stopwatch.getUserTime() << " ms");
		BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/size << " ms" );

		// meassure find_close
		stopwatch.start();
		for(size_type i=0; i < size; ++i){
			if(v[tc][i]==1)
				k += bp_support.find_close(i);
			else
				k -= 1;
		}
		stopwatch.stop();
		BOOST_MESSAGE("Time for "<<size/2<<" find_close calls "<< stopwatch.getUserTime() << " ms");
		BOOST_MESSAGE("Per call: "<< 2*stopwatch.getUserTime()/size << " ms" );

		// meassure find_close
		stopwatch.start();
		for(size_type i=0; i < size; ++i){
			if(v[tc][i]==0)
				k += bp_support.find_open(i);
			else
				k -= 1;
		}
		stopwatch.stop();
		BOOST_MESSAGE("Time for "<< size/2 <<" find_open calls "<< stopwatch.getUserTime() << " ms");
		BOOST_MESSAGE("Per call: "<< 2*stopwatch.getUserTime()/size << " ms" );

		// meassure enclose 
		stopwatch.start();
		for(size_type i=0; i < size; ++i){
			if(v[tc][i]==1)
				k += bp_support.enclose(i);
			else
				k -= 1;
		}
		stopwatch.stop();
		BOOST_MESSAGE("Time for "<< size/2 <<" enclose calls "<< stopwatch.getUserTime() << " ms");
		BOOST_MESSAGE("Per call: "<< 2*stopwatch.getUserTime()/size << " ms" );

		// meassure restricted_enclose
		std::vector<size_type> i_and_j(size*2,0);
		// generate one valid test
		bool found_restricted_enclose_tc = false; 
		for(size_type i=1; i < size and !found_restricted_enclose_tc; ++i){
			if( v[tc][i-1] == 0 and v[tc][i] == 1 ){
				found_restricted_enclose_tc = true;
				i_and_j[0] = bp_support.find_open(i-1);
				i_and_j[1] = i;
			}
		}
		if(found_restricted_enclose_tc){
			// generate test cases
			for(size_type i=1; i < size; ++i){
				size_type ii = bp_support.select(1 + rand() % (size/2));
				size_type mii = bp_support.find_close(ii);
				size_type ones = bp_support.rank(mii);
				if( ones < size/2 ){
					size_type jj = bp_support.select( ones + 1 + rand() % (size/2-ones) );
					i_and_j[2*i]   = ii;
					i_and_j[2*i+1] = jj; 
				}else{
					i_and_j[2*i]   = i_and_j[2*i-2];
					i_and_j[2*i+1] = i_and_j[2*i-1];
				}
			}
			stopwatch.start();
			for(size_type i=0; i < size; ++i){
				k += bp_support.restricted_enclose(i_and_j[i<<1], i_and_j[(i<<1)+1]);
			}
			stopwatch.stop();
			BOOST_MESSAGE("Time for "<< size <<" random chosen restricted_enclose calls "<< stopwatch.getUserTime() << " ms");
			BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/size << " ms" );
		}else{ // found no valid test case
			BOOST_MESSAGE("Could not find test cases for restricted_enclose!");
		}
	}

	BOOST_MESSAGE("k="<<k);
}

BOOST_AUTO_TEST_SUITE_END()
