#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MAIN 
#include <boost/test/unit_test.hpp>
#include <sdsl/vectors.hpp>
#include <sdsl/testutils.hpp> 
#include <ctime> // for performance measurement

using namespace boost::unit_test;
using namespace boost::test_tools;
using namespace sdsl;

BOOST_AUTO_TEST_SUITE( coder_performance_test_suite )

BOOST_AUTO_TEST_CASE( performance_test ){
	unit_test_log.set_threshold_level( log_messages );
	BOOST_MESSAGE( "Running performance test for coders..." );
	int_vector<> v,w,x,y,z;
	uint8_t intWidth	= 33;
	size_t  size		= 10000000;
	v.setIntWidth(intWidth);
	v.resize( size );

	for(size_t i=0; i< v.size(); ++i){
		v[i] = (i&bit_magic::Li1Mask[intWidth])+1;
	}
	
	stop_watch stopwatch;

	stopwatch.start();
	coder::fibonacci::encode(v,z);
	stopwatch.stop();
	BOOST_MESSAGE("Fibonacci encoding of vector v (size = "<<v.bit_size()<<" bits) takes "<< stopwatch.getUserTime() << " ms" );

	stopwatch.start();
	coder::fibonacci::decode(z,w);
	stopwatch.stop();

	BOOST_MESSAGE("Fibonacci decoding of vector z (size = "<<z.bit_size()<<" bits) takes "<< stopwatch.getUserTime() << " ms" );

	BOOST_CHECK( v == w );// BOOST_CHECK_EQUAL( v, w) is very slow... why?

	stopwatch.start();
	coder::elias_delta::encode(v,z);
	stopwatch.stop();
	BOOST_MESSAGE("Elias delta encoding of vector v (size = "<<v.bit_size()<<" bits) takes "<< stopwatch.getUserTime() << " ms" );

	stopwatch.start();
	coder::elias_delta::decode(z,w);
	stopwatch.stop();

	BOOST_MESSAGE("Elias delta decoding of vector z (size = "<<z.bit_size()<<" bits) takes "<< stopwatch.getUserTime() << " ms" );

	BOOST_CHECK( v == w );// BOOST_CHECK_EQUAL( v, w) is very slow... why?

	stopwatch.start();
	coder::ternary::encode(v,z);
	stopwatch.stop();
	BOOST_MESSAGE("Ternary encoding of vector v (size = "<<v.bit_size()<<" bits) takes "<< stopwatch.getUserTime() << " ms" );

	stopwatch.start();
	coder::ternary::decode(z,w);
	stopwatch.stop();

	BOOST_MESSAGE("Ternary decoding of vector z (size = "<<z.bit_size()<<" bits) takes "<< stopwatch.getUserTime() << " ms" );

	BOOST_CHECK( v == w );// BOOST_CHECK_EQUAL( v, w) is very slow... why?
	if(v!=w){
		for(int_vector<>::size_type i=0; i<v.size();++i)
			if(v[i]!=w[i]){
				BOOST_MESSAGE("i="<<i<<" v[i]="<<v[i]<<"  w[i]="<<w[i]<<" "<<*(z.data()) );
				for(int_vector<>::size_type j=i+1; j<v.size() && j-i < 10;++j)
					BOOST_MESSAGE("j="<<j<<" v[j]="<<v[j]<<"  w[j]="<<w[j]<<" "<<*(z.data()) );
				return;
			}
	}	

	size_t n = 10000000000ULL; // be able to allocate 1GB
	size = 10000000;
	intWidth = bit_magic::l1BP(n)+1;
	v.setIntWidth( intWidth );
	v.resize( size );
	for(size_t i=0; i < v.size(); ++i){
		v[i] = (n/size)*(i+1);
	}	
	w.setIntWidth( intWidth );
	w.resize( size );
	w[0] = v[0];
	for(size_t i=1; i < v.size(); ++i){
		w[i] = v[i] - v[i-1];
	}
	
	stopwatch.start();
	coder::fibonacci::encode(w,z);
	stopwatch.stop();
	BOOST_MESSAGE("Fibonacci encoding of vector w (size = "<<w.bit_size()<<" bits) takes "<< stopwatch.getUserTime() << " ms" );

	stopwatch.start();
	coder::fibonacci::decode(z,x);
	stopwatch.stop();

	BOOST_MESSAGE("Fibonacci decoding of vector z (size = "<<z.bit_size()<<" bits) takes "<< stopwatch.getUserTime() << " ms" );

	BOOST_CHECK( w == x );

	stopwatch.start();
	coder::elias_delta::encode(w,z);
	stopwatch.stop();
	BOOST_MESSAGE("Elias delta encoding of vector w (size = "<<w.bit_size()<<" bits) takes "<< stopwatch.getUserTime() << " ms" );

	stopwatch.start();
	coder::elias_delta::decode(z,x);
	stopwatch.stop();

	BOOST_MESSAGE("Elias delta decoding of vector z (size = "<<z.bit_size()<<" bits) takes "<< stopwatch.getUserTime() << " ms" );

	BOOST_CHECK( w == x );
	
	stopwatch.start();
	coder::ternary::encode(w,z);
	stopwatch.stop();
	BOOST_MESSAGE("Ternary encoding of vector w (size = "<<w.bit_size()<<" bits) takes "<< stopwatch.getUserTime() << " ms" );

	stopwatch.start();
	coder::ternary::decode(z,x);
	stopwatch.stop();

	BOOST_MESSAGE("Ternary decoding of vector z (size = "<<z.bit_size()<<" bits) takes "<< stopwatch.getUserTime() << " ms" );

	BOOST_CHECK( w == x );


	size_t diff = 5;
	n = 100000000;
	v.setIntWidth( bit_magic::l1BP( n*(diff) )+1 );
	v.resize( n );
	v[0] = (lrand48()%diff)+1;
	for(size_t i=1; i < v.size(); ++i){
		v[i] = v[i-1] + (lrand48()%diff) + 1;
	}
	w.setIntWidth( v.getIntWidth() );
	w.resize( v.size() );
	w[0] = v[0];
	for(size_t i=1; i < v.size(); ++i){
		w[i] = v[i]-v[i-1];
//		BOOST_MESSAGE(" "<<(int)w[i]);
	}

	stopwatch.start();
	coder::fibonacci::encode(w,z);
	stopwatch.stop();
	BOOST_MESSAGE("Fibonacci encoding of vector w (size = "<<w.bit_size()<<" bits) takes "<< stopwatch.getUserTime() << " ms" );

	stopwatch.start();
	coder::fibonacci::decode(z,x);
	stopwatch.stop();

	BOOST_MESSAGE("Fibonacci decoding of vector z (size = "<<z.bit_size()<<" bits) takes "<< stopwatch.getUserTime() << " ms" );

	BOOST_CHECK( w.bit_size() == x.bit_size() );
	BOOST_CHECK( w.getIntWidth() == x.getIntWidth() );
	
	BOOST_CHECK( w == x );

	stopwatch.start();
	coder::elias_delta::encode(w,z);
	stopwatch.stop();
	BOOST_MESSAGE("Elias delta encoding of vector w (size = "<<w.bit_size()<<" bits) takes "<< stopwatch.getUserTime() << " ms" );

	stopwatch.start();
	coder::elias_delta::decode(z,x);
	stopwatch.stop();

	BOOST_MESSAGE("Elias delta decoding of vector z (size = "<<z.bit_size()<<" bits) takes "<< stopwatch.getUserTime() << " ms" );

	BOOST_CHECK( w == x );

	stopwatch.start();
	coder::ternary::encode(w,z);
	stopwatch.stop();
	BOOST_MESSAGE("Ternary encoding of vector w (size = "<<w.bit_size()<<" bits) takes "<< stopwatch.getUserTime() << " ms" );

	stopwatch.start();
	coder::ternary::decode(z,x);
	stopwatch.stop();

	BOOST_MESSAGE("Ternary decoding of vector z (size = "<<z.bit_size()<<" bits) takes "<< stopwatch.getUserTime() << " ms" );

	BOOST_CHECK( w == x );

}

BOOST_AUTO_TEST_SUITE_END()
