#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MAIN 
#include <boost/test/unit_test.hpp>
#include <sdsl/vectors.hpp>
#include <sdsl/enc_vector_prac.hpp>
#include <sdsl/testutils.hpp>
#include <vector>
#include <fstream>
#include <sstream>
#if BOOST_WORKAROUND( __GNUC__, < 3)
typedef boost::test_tools::output_test_stream onullstream_type;
#else
#include <boost/test/utils/nullstream.hpp>
typedef boost::onullstream onullstream_type;
#endif

using namespace boost::unit_test;
using namespace sdsl;

typedef std::vector<uint64_t> tVI;
typedef enc_vector<coder::fibonacci> tEncV;
typedef enc_vector<coder::elias_delta> tEncVE;
typedef enc_vector<coder::ternary> tEncVT;
typedef enc_vector_prac<coder::fibonacci, 16> tEncVP;

BOOST_AUTO_TEST_SUITE( enc_vector_performance_suite )


BOOST_AUTO_TEST_CASE( coder_performance ){
	onullstream_type nullstream;
	unit_test_log.set_threshold_level( log_messages );
	stop_watch stopwatch;
	tVI a(1000000,0);
	a[0] = 0;
	for(size_t i=1;i < a.size();++i){
		a[i]= a[i-1]+rand()%5+1;
	}
	
	stopwatch.start();
	tEncV encV(a); 
	stopwatch.stop();
	BOOST_MESSAGE("Build of Fibonacci EncVector takes "<<stopwatch.getUserTime()<<" ms");
	BOOST_MESSAGE("Size bytes "<< encV.serialize(nullstream));


	enc_vector<>::value_type k1=0,k2=0,k3=0,k4=0;
	stopwatch.start();
	for(tEncV::const_iterator it = encV.begin(), end = encV.end(); it != end; ++it){
		k1 += *it;
	}
	stopwatch.stop();
	BOOST_MESSAGE("Iteration of Fibonacci EncVector takes "<<stopwatch.getUserTime()<<" ms");
	stopwatch.start();
	for(enc_vector<>::size_type i=0, size = encV.size(); i<size; ++i){
		k2 += encV[i];
	}
	stopwatch.stop();
	BOOST_MESSAGE("Random Access Iteration of Fibonacci EncVector takes "<<stopwatch.getUserTime()<<" ms");
	BOOST_CHECK(k1 == k2);
	k2=0;
	{// clear encV
		tEncV temp;
		temp.swap(encV);
	}

	{
		stopwatch.start();
		tEncVP encVP(a);
		stopwatch.stop();
		BOOST_MESSAGE("Build of Fibonacci EncVectorPrac takes "<<stopwatch.getUserTime()<<" ms");
		BOOST_MESSAGE("Size bytes "<< encVP.serialize(nullstream));
		k2=0;
		stopwatch.start();
		for(int i=0, size = encVP.size(); i<size; ++i){
			k2 += encVP[i];
		}
		stopwatch.stop();
		BOOST_MESSAGE("Random Access Iteration of Fibonacci EncVectorPrac takes "<<stopwatch.getUserTime()<<" ms");
		BOOST_CHECK(k1==k2);
	}	

	stopwatch.start();
	tEncVE encVE(a); 
	stopwatch.stop();
	BOOST_MESSAGE("Build of EliasDelta EncVector takes "<<stopwatch.getUserTime()<<" ms");
	BOOST_MESSAGE("Size bytes "<< encVE.serialize(nullstream));

	k2=0;
	stopwatch.start();
	for(tEncVE::const_iterator it = encVE.begin(), end = encVE.end(); it != end; ++it){
		k2 += *it;
	}
	stopwatch.stop();
	BOOST_MESSAGE("Iteration of EliasDelta EncVector takes "<<stopwatch.getUserTime()<<" ms");
	stopwatch.start();
	for(enc_vector<>::size_type i=0, size = encVE.size(); i<size; ++i){
		k3 += encVE[i];
	}
	stopwatch.stop();
	BOOST_MESSAGE("Random Access Iteration of EliasDelta EncVector takes "<<stopwatch.getUserTime()<<" ms");
	BOOST_CHECK(k2 == k3);
	BOOST_CHECK(k1 == k2);
	k3=0;
	{// clear encV
		tEncVE temp;
		temp.swap(encVE);
	}

	stopwatch.start();
	tEncVT encVT(a); 
	stopwatch.stop();
	BOOST_MESSAGE("Build of Ternary EncVector takes "<<stopwatch.getUserTime()<<" ms");
	BOOST_MESSAGE("Size bytes "<< encVT.serialize(nullstream));

	stopwatch.start();
	for(tEncVT::const_iterator it = encVT.begin(), end = encVT.end(); it != end; ++it){
		k3 += *it;
	}
	stopwatch.stop();
	BOOST_MESSAGE("Iteration of Ternary EncVector takes "<<stopwatch.getUserTime()<<" ms");
	stopwatch.start();
	for(enc_vector<>::size_type i=0, size = encVT.size(); i<size; ++i){
		k4 += encVT[i];
	}
	stopwatch.stop();
	BOOST_MESSAGE("Random Access Iteration of Ternary EncVector takes "<<stopwatch.getUserTime()<<" ms");
	BOOST_CHECK(k3 == k4);
	BOOST_CHECK(k1 == k3);
	k3=0;
	{// clear encV
		tEncVT temp;
		temp.swap(encVT);
	}
}

BOOST_AUTO_TEST_CASE( performance_test )
{
	onullstream_type nullstream;
	unit_test_log.set_threshold_level( log_messages );
	BOOST_TEST_MESSAGE("Running performance test for enc_vector");
	stop_watch stopwatch; // stopwatch for performance measurement
	size_t n;

	const size_t kase = 12;
	uint64_t ns[kase] = {800000,1600000,3200000,6400000,12800000,25600000,51200000,102400000,204800000,409600000,614400000,819200000};
	tVI test_size(kase);
	for(size_t i=0;i<kase;++i) test_size[i] = ns[i];

	BOOST_MESSAGE("Vector size  Random Access Interation Time of    ");


	std::stringstream log;
	log<<"# n (number of elements)  size of stl vector    usr usr+sys real (seq acc)  usr usr+sys real (rand acc) \
	   	  of EncVector usr usr+sys real (seq acc) usr usr+sys real (rand acc)\n\
# times in ms = 10^-3 sec \n\
# sizes in bytes"<<std::endl;
	for(size_t tc = 0; tc<test_size.size() and false; ++tc){
		log<<test_size[tc]<<" ";
		BOOST_MESSAGE("Number of elements n = "<<test_size[tc]);
		uint64_t t=0, t1 = 0; //check sum
		// first test stl vector
		tVI a(test_size[tc],0);
		n=a.size();
		a[0] = 0;
		for(size_t i=1;i < a.size();++i)
			a[i]= a[i-1] + rand()%100;

		BOOST_MESSAGE("size of the STL vector in bytes "<<a.size()*sizeof(a[0]));
		log<< a.size()*sizeof(a[0]) << " ";
		
		t=0;
		BOOST_MESSAGE("Performing seuqential acccess for the STL vector:");
		stopwatch.start();
		for(tVI::const_iterator it=a.begin(), end=a.end(); it!=end; ++it){
			t += *it;
		}
		stopwatch.stop();
		BOOST_MESSAGE("usr time  "<< stopwatch.getUserTime() << " " );
		BOOST_MESSAGE("usr+sys time  "<< stopwatch.getUserTime() + stopwatch.getSysTime() << " " );
		BOOST_MESSAGE("real time: "<< stopwatch.getRealTime() );
		log<<stopwatch.getUserTime()<<" "<<stopwatch.getUserTime()+stopwatch.getSysTime()<<" "<<stopwatch.getRealTime()<<" ";

		t1 = 0;
		BOOST_MESSAGE("Performing random acccess for the STL vector:");
		stopwatch.start();
		for(size_t i=0; i!=n;++i){
			t1 += a[i&1 ? n - i : i];
		}
		stopwatch.stop();
		BOOST_MESSAGE("usr time  "<< stopwatch.getUserTime() << " " );
		BOOST_MESSAGE("usr+sys time  "<< stopwatch.getUserTime() + stopwatch.getSysTime() << " " );
		BOOST_MESSAGE("real time: "<< stopwatch.getRealTime() );
		log<<stopwatch.getUserTime()<<" "<<stopwatch.getUserTime()+stopwatch.getSysTime()<<" "<<stopwatch.getRealTime()<<" ";
		BOOST_CHECK(t == t1);

		tEncV v(a); // create encoded vector

		size_t v_byte_size = v.serialize(nullstream);
		BOOST_MESSAGE("size of the EncVector in bytes "<<v_byte_size);
		log<< v_byte_size << " ";
		{// free memory of stl vector a
			tVI b;
			a.swap(b);
		}

		t1=0;
		BOOST_MESSAGE("Performing sequential acccess for the EncVector vector:");
		stopwatch.start();
		for(tEncV::const_iterator it=v.begin(), end = v.end(); it != end; ++it){
			t1 += *it;
		}
		stopwatch.stop();
		BOOST_MESSAGE("usr time "<< stopwatch.getUserTime() << " " );
		BOOST_MESSAGE("usr+sys time  "<< stopwatch.getUserTime()+stopwatch.getSysTime() << " " );
		BOOST_MESSAGE("real time: "<< stopwatch.getRealTime() );
		log<<stopwatch.getUserTime()<<" "<<stopwatch.getUserTime()+stopwatch.getSysTime()<<" "<<stopwatch.getRealTime()<<" ";
		BOOST_CHECK(t==t1);

		t1=0;
		BOOST_MESSAGE("Performing random acccess for the EncVector vector:");
		stopwatch.start();
		for(size_t i=0; i<n;++i){
			t1 += v[i&1 ? n - i : i];
		}
		stopwatch.stop();
		BOOST_MESSAGE("usr time "<< stopwatch.getUserTime() << " " );
		BOOST_MESSAGE("usr+sys time  "<< stopwatch.getUserTime()+stopwatch.getSysTime() << " " );
		BOOST_MESSAGE("real time: "<< stopwatch.getRealTime() );
		log<<stopwatch.getUserTime()<<" "<<stopwatch.getUserTime()+stopwatch.getSysTime()<<" "<<stopwatch.getRealTime()<<" ";
		BOOST_CHECK(t==t1);
		
		log<<std::endl;
	}
	std::cerr<<log.str()<<std::endl;
	std::ofstream 	logfile;
	std::string file_name= "enc_vector_performance_"+sdsl::clock::getTimeString()+".log";
	logfile.open(file_name.c_str(), std::ios::out);
	if(!logfile){
		BOOST_MESSAGE("Could not write log file.");
	}else{
		logfile << log.str() << std::endl;
		logfile.close();
		BOOST_MESSAGE("wrote logfile: " << file_name );
	}


}

BOOST_AUTO_TEST_SUITE_END()

/*test_suite*
init_unit_test_suite( int , char* [] )
{
    framework::master_test_suite().p_name.value = "SDSBitVector Test Suite";

    // with explicit registration we could specify a test case timeout
    framework::master_test_suite().add( BOOST_TEST_CASE( &operator_test ), 0 );

    return 0; 
}
*/
