#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MAIN 
#include <boost/test/unit_test.hpp>
#include <sdsl/vectors.hpp>
#include <sdsl/testutils.hpp>
#include <algorithm> // for stl conform test 

using namespace boost::unit_test;
using namespace sdsl;

BOOST_AUTO_TEST_SUITE( int_vector_performance_suite )

BOOST_AUTO_TEST_CASE( performance_test )
{
	unit_test_log.set_threshold_level( log_messages );
	BOOST_TEST_MESSAGE("Running performance test for SDSBitVector");
	
	size_t s = 100000000;
	
	stop_watch stopWatch;
	stopWatch.start();
	int_vector<> v(s, 0, 3);
	stopWatch.stop();
	
	BOOST_MESSAGE("Construction of int_vector v (size = "<<v.bit_size()<<" bits) takes "<< stopWatch.getUserTime() << " ms" );

	BOOST_CHECK_EQUAL( v.size(), s );

	stopWatch.start();
	util::setOneBits(v);
	stopWatch.stop();

	BOOST_MESSAGE("Set all bits in v (size = "<<v.bit_size()<<" bits) to 1 takes "<< stopWatch.getUserTime() << " ms" );

	uint64_t sum = 0;
	stopWatch.start();
	for(int_vector<>::iterator it=v.begin(); it!=v.end();++it){
		sum += *it;
	}
	stopWatch.stop();
	BOOST_CHECK_EQUAL( v.size()*7, (size_t)sum );
	BOOST_MESSAGE("Iterating through all elements in v (size = "<<v.bit_size()<<" bits) with iterator takes "<< stopWatch.getUserTime() << " ms" );
	sum = 0;
	stopWatch.start();
	for(size_t i=0; i < v.size() ;++i){
		sum += v[i];
	}
	stopWatch.stop();
	BOOST_CHECK_EQUAL( v.size()*7, (size_t)sum );
	BOOST_MESSAGE("Iterating through all elements in v (size = "<<v.bit_size()<<" bits) with loop and access through [] "<< stopWatch.getUserTime() << " ms" );

	sum = 0;
	stopWatch.start();
	for(int_vector<>::const_iterator it = v.begin(); it != v.end();++it){
		sum += *it;
//		if(  it == v.begin() ){
//			BOOST_MESSAGE("  diff "<<(v.end() - it) );
//		}
	}
	stopWatch.stop();
	BOOST_CHECK_EQUAL( v.size()*7, (size_t)sum );
	BOOST_MESSAGE("Iterating through all elements in v (size = "<<v.bit_size()<<" bits) with const iterator takes "<< stopWatch.getUserTime() << " ms" );

	v.resize(0);
	v.setIntWidth(64);
	v.resize(10000000);
	BOOST_MESSAGE("Set v to 64 bit int width and resize it to "<<v.size()<<" elements");
	uint64_t k1=0,k2=0;	
	stopWatch.start();
	for(int_vector<>::size_type i=0, size=v.size(); i<size; ++i){
		v[i] = i;	
	}
	stopWatch.stop();
	BOOST_MESSAGE("Initializing all values in v takes "<< stopWatch.getUserTime() <<" ms");

	stopWatch.start();
	for(int_vector<>::size_type i=0, size=v.size(); i<size; ++i){
		k1 += v[i];	
	}
	stopWatch.stop();
	BOOST_MESSAGE("Summing up all values in v takes "<< stopWatch.getUserTime() <<" ms");
	int_vector<64> v2(v.size());
	for(int_vector<>::size_type i=0, size=v.size(); i<size; ++i){
		v2[i] = v[i];
//		BOOST_CHECK( v2[i] == v[i] );
	}
	stopWatch.start();
	for(int_vector<>::size_type i=0, size=v2.size(); i<size; ++i){
		v2[i] = i;	
	}
	stopWatch.stop();
	BOOST_MESSAGE("Initializing all values in v takes "<< stopWatch.getUserTime() <<" ms");

	const int_vector<64> &v2r = v2;
	stopWatch.start();
	for(int_vector<>::size_type i=0, size=v2.size(); i<size; ++i){
		k2 += v2r[i];	
	}
	stopWatch.stop();
	BOOST_MESSAGE("Summing up all values in v2 takes "<< stopWatch.getUserTime() <<" ms");
	BOOST_CHECK(k1==k2);	
	v.resize(0); v2.resize(0);

	bit_vector b1(10000000);
	int_vector<> b2(b1.size(),0,1);
	for(int_vector<>::size_type i=0;i<b1.size();++i){
		b1[i] = b2[i] = rand()&1;
	}	
	const bit_vector &br1 = b1;
	const int_vector<> &br2 = b2;
	k1 = k2 = 0;
	stopWatch.start();
	for(int_vector<>::size_type i=0, size=b1.size(); i<size; ++i){
		k1 += br1[i];	
	}
	stopWatch.stop();
	BOOST_MESSAGE("Summing up all values in b1 takes "<< stopWatch.getUserTime() <<" ms");

	stopWatch.start();
	for(int_vector<>::size_type i=0, size=b1.size(); i<size; ++i){
		k1 += b1[i];	
	}
	stopWatch.stop();
	BOOST_MESSAGE("Summing up all values in b1 takes "<< stopWatch.getUserTime() <<" ms");


	stopWatch.start();
	for(int_vector<>::size_type i=0, size=b1.size(); i<size; ++i){
		k2 += br2[i];	
	}
	stopWatch.stop();
	BOOST_MESSAGE("Summing up all values in b2 takes "<< stopWatch.getUserTime() <<" ms");

	stopWatch.start();
	for(int_vector<>::size_type i=0, size=b1.size(); i<size; ++i){
		k2 += b2[i];	
	}
	stopWatch.stop();
	BOOST_MESSAGE("Summing up all values in b2 takes "<< stopWatch.getUserTime() <<" ms");
	BOOST_CHECK(k1==k2);
	
}

BOOST_AUTO_TEST_SUITE_END()

/*test_suite*
init_unit_test_suite( int , char* [] )
{
    framework::master_test_suite().p_name.value = "SDSBitVector Test Suite";

    // with explicit registration we could specify a test case timeout
    framework::master_test_suite().add( BOOST_TEST_CASE( &operator_test ), 0 );

    return 0; 
}
*/
