/* sdsl - succinct data structures library
    Copyright (C) 2008 Simon Gog 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/ .
*/
/*! \file coder.hpp
    \brief coder.hpp contains the coder namespace and sdsl::coder::fibonacci, sdsl::coder::elias_delta, sdsl::coder::ternary, and sdsl::coder::run_length
	\author Simon Gog
 */
#ifndef SDSL_CODER
#define SDSL_CODER

#include "int_vector.hpp"

namespace sdsl{

//! Namespace for the different coder of the sdsl.
namespace coder{

template<class Coder>
class run_length{
	public:
		typedef uint64_t size_type;
		static void encode(uint64_t x, uint64_t* &z, uint8_t offset);
		static uint64_t encoding_length(const uint64_t *s, uint8_t s_offset, size_type bit_length);
};

template<class Coder>
typename run_length<Coder>::size_type run_length<Coder>::encoding_length(const uint64_t *s, uint8_t s_offset, size_type bit_length){
	assert(s_offset < 64);
	size_type i=0;
	uint64_t w = (*s >> s_offset);
	uint8_t last_bit = w&1;
	size_type result = 0;
	while(i < bit_length){
		size_type len = 0;	
		while( last_bit == (w&1) and  i < bit_length ){
//			std::cout<<w<<" "<<i<<std::endl;
			++len; ++i; ++s_offset;
			w >>= 1;
			if( s_offset == 64 ){ s_offset = 0; w = *(++s); }
		}	
//		std::cout<<"len="<<Coder::encoding_length(len)<<std::endl;
		last_bit = (w&1);
		result += Coder::encoding_length(len);
	}
	return result;
}	

//! A class to encode and decode between Fibonacci and binary code.  
class fibonacci{
	public:
		typedef uint64_t size_type;

		static const uint8_t min_codeword_length = 2; // 11 represents 1 and is the code word with minimum length
		//! Get the number of bits that are necessary to encode the value w in fibonacci code.
		/*! \param w 64bit integer to get the length of its fibonacci encoding. Inclusive the terminating 1 of the code.
		 */
		static uint8_t encoding_length(uint64_t w);
		//! Decode n Fibonacci encoded bits beginning at start_idx in the bitstring "data"
		/* \param data Bitstring
		   \param start_idx Starting index of the decoding.
		   \param n Number of values to decode from the bitstring.
		   \param it Iterator
		 */
		template<bool sumup, bool increment, class Iterator>
		static uint64_t decode(const uint64_t *data, const size_type start_idx, size_type n, Iterator it=(Iterator)NULL);

		template<bool sumup, bool increment, class Iterator>
		static uint64_t decode1(const uint64_t *data, const size_type start_idx, size_type n, Iterator it=(Iterator)NULL);



		//! Decode n Fibonacci encoded integers beginning at start_idx in the bitstring "data"  and return the sum of these values.
	    /*! \param data Pointer to the beginning of the Fibonacci encoded bitstring.
		    \param start_idx Index of the first bit to endcode the values from.
			\param n Number of values to decode from the bitstring. Attention: There have to be at least n encoded values in the bitstring.
	     */
		static uint64_t decode_prefix_sum(const uint64_t *data, const size_type start_idx, size_type n);

		//! Deocde n Fibonacci encoded integers beginning at start_idx and ending at end_idx (exclusive) in the bitstring "data" and return the sum of these values.
		/*! \sa decode_prefix_sum 
		  */
		static uint64_t decode_prefix_sum(const uint64_t *data, const size_type start_idx, const size_type end_idx, size_type n);

		template<class int_vector1, class int_vector2>
		static bool encode(const int_vector1 &v, int_vector2 &z);

		template<class int_vector>
		static uint64_t* raw_data(int_vector &v){
			return v.m_data;
		};

		//! Encode one positive integer x to an int_vector at bit position start_idx.
		/*! \param x Positive integer to encode. 
		    \param z Raw data of vector to write the encoded form of x.
			\param offset Start offset to write the encoded form of x in z. \f$0\leq offset< 64\f$.
		 */ 
		static void encode(uint64_t x, uint64_t* &z, uint8_t &offset);	

		template<class int_vector1, class int_vector2>
		static bool decode(const int_vector1 &z, int_vector2 &v);
};

//! A class to encode and decode between ternary and binary code.
class ternary{
public: 
		typedef uint64_t size_type;

		static const uint8_t min_codeword_length = 4; // 11 01 represents 1 and is the code word with minumum length

		static uint8_t encoding_length(uint64_t w);

		template<bool sumup, bool increment, class Iterator>
		static uint64_t decode(const uint64_t *data, const size_type start_idx, size_type n, Iterator it=(Iterator)NULL);

		static uint64_t decode_prefix_sum(const uint64_t *data, const size_type start_idx, size_type n);
		static uint64_t decode_prefix_sum(const uint64_t *data, const size_type start_idx, const size_type end_idx, size_type n);

		template<class int_vector>
		static bool encode(const int_vector &v, int_vector &z);
		template<class int_vector>
		static bool decode(const int_vector &z, int_vector &v);

		//! Encode one positive integer x to an int_vector at bit position start_idx.
		/* \param x Positive integer to encode.
		   \param z Raw data of vector to write the encoded form of x.
		   \param start_idx Beginning bit index to write the encoded form ox x in z.
		*/
		static void encode(uint64_t x, uint64_t *&z, uint8_t &offset);

		template<class int_vector>
		static uint64_t* raw_data(int_vector &v){
			return v.m_data;
		};	
};

//! A class to encode and decode between Elias-\f$\delta\f$ and binary code.
class elias_delta{
public:
		typedef uint64_t size_type;

		static const uint8_t min_codeword_length = 1; // 1 represents 1 and is the code word with minimum length
		static uint8_t encoding_length(uint64_t);
		//! Decode n Elias-delta encoded bits beginning at start_idx in the bitstring "data"
		/* \param data Bitstring
		   \param start_idx Starting index of the decoding.
		   \param n Number of values to decode from the bitstring.
		   \param it Iterator to decode the values.
		 */
		template<bool sumup, bool increment,class Iterator>
		static uint64_t decode(const uint64_t *data, const size_type start_idx, size_type n, Iterator it=(Iterator)NULL);

		//! Decode n EliasDelta encoded integers beginning at start_idx in the bitstring "data"  and return the sum of these values.
	    /*! \param data Pointer to the beginning of the EliasDelta encoded bitstring.
		    \param start_idx Index of the first bit to endcode the values from.
			\param n Number of values to decode from the bitstring. Attention: There have to be at least n encoded values in the bitstring.
	     */
		static uint64_t decode_prefix_sum(const uint64_t *data, const size_type start_idx, size_type n);
		static uint64_t decode_prefix_sum(const uint64_t *data, const size_type start_idx, const size_type end_idx, size_type n);

		template<class int_vector>
		static bool encode(const int_vector &v, int_vector &z);
		template<class int_vector>
		static bool decode(const int_vector &z, int_vector &v);

		//! Encode one positive integer x to an int_vector at bit position start_idx.
		/* \param x Positive integer to encode.
		   \param z Raw data of vector to write the encoded form of x.
		   \param start_idx Beginning bit index to write the encoded form ox x in z.
		*/
		static void encode(uint64_t x, uint64_t *&z, uint8_t &offset);

		template<class int_vector>
		static uint64_t* raw_data(int_vector &v){
			return v.m_data;
		};	
};


inline uint8_t fibonacci::encoding_length(uint64_t w){
	// This limit for the leftmost 1bit in the resulting fib code could be improved using a table
	uint8_t len_1 = bit_magic::l1BP(w); // len-1 of the fib code
	while( ++len_1 < (uint8_t)(sizeof(bit_magic::Fib)/sizeof(bit_magic::Fib[0])) && w >= bit_magic::Fib[len_1] );
	return len_1+1;
}

// \sa coder::fibonacci::encoding_length
inline uint8_t elias_delta::encoding_length(uint64_t w){
	uint8_t len_1 = bit_magic::l1BP(w);
   	return len_1 + (bit_magic::l1BP(len_1+1)<<1) + 1;
}

inline uint8_t ternary::encoding_length(uint64_t w){
	if(w==0) return 2;
	uint8_t res = 2;   // add two for the terminating "11" bit pair
	do{
		w/=3;
		res+=2;
	}while(w!=0);
	return res;
}

template<class int_vector1, class int_vector2>
inline bool fibonacci::encode(const int_vector1 &v, int_vector2 &z){
	uint64_t z_bit_size = 0;
	for(typename int_vector1::const_iterator it=v.begin(), end = v.end(); it != end; ++it){
		if( *it == 0 ){
			throw std::logic_error("fibonacci::encode(const SDSBitVector &v, SDSBitVector &z); entry of v equals 0 that cannot be encoded!");
		}
		z_bit_size += encoding_length(*it);
	}
	z.bit_resize( z_bit_size ); 
	if( z_bit_size & 0x3F ){ // if z_bit_size % 64 != 0
		*(z.m_data + (z_bit_size>>6)) = 0; // initialize last word
	}
	uint64_t *z_data 	= z.m_data;
	uint8_t offset 		= 0;
	register uint64_t w;
   	register uint64_t fibword_high = 0x0000000000000001ULL, fibword_low;
	register uint64_t t;	
	for(typename int_vector1::const_iterator it=v.begin(), end = v.end(); it != end; ++it){
		w = *it;
		int8_t len_1 = encoding_length(w)-1,j;
		fibword_low = 0x0000000000000001ULL;

		if( len_1 >= 64 ){ // length > 65
			fibword_high = 0x0000000000000001ULL;
			for(j = len_1-1; j>63; --j){
				fibword_high <<= 1;
				if( w >= (t=bit_magic::Fib[j]) ){
					w -= t;
					fibword_high |= 1;
					if( w and j>64 ){
						fibword_high <<= 1;
						--j;
					}else{
						fibword_high <<= (64-j);
						break;
					}
				}
			}
			j		= 64;
		}
		else{
			j = len_1-1;
		}

		for( ; j >= 0; --j){
			fibword_low <<= 1;
			if( w >= (t=bit_magic::Fib[j]) ){
				w -= t;
				fibword_low |= 1;
				if( w ){
					fibword_low <<= 1;
					--j;
				}else{
					fibword_low <<= (j);
					break;
				}
			}	
		}
		if( len_1 >=64 ){
			bit_magic::writeIntAndMove(z_data, fibword_low, offset, 64);
			bit_magic::writeIntAndMove(z_data, fibword_high, offset, len_1 - 63);
		}else{
			bit_magic::writeIntAndMove(z_data, fibword_low, offset, (len_1&0x3F) +1);
		}
	}
	z.setIntWidth(v.getIntWidth());
	return true;	
}

inline void fibonacci::encode(uint64_t x, uint64_t* &z, uint8_t &offset){
   	register uint64_t fibword_high = 0x0000000000000001ULL, fibword_low;
	register uint64_t t;	
	int8_t len_1 = encoding_length(x)-1,j;
	fibword_low = 0x0000000000000001ULL;

	if( len_1 >= 64 ){ // length > 65
		fibword_high = 0x0000000000000001ULL;
		for(j = len_1-1; j>63; --j){
			fibword_high <<= 1;
			if( x >= (t=bit_magic::Fib[j]) ){
				x -= t;
				fibword_high |= 1;
				if( x and j>64 ){
					fibword_high <<= 1;
					--j;
				}else{
					fibword_high <<= (64-j);
					break;
				}
			}
		}
		j	= 64;
	}
	else{
		j = len_1-1;
	}
	for( ; j >= 0; --j){
		fibword_low <<= 1;
		if( x >= (t=bit_magic::Fib[j]) ){
			x -= t;
			fibword_low |= 1;
			if( x ){
				fibword_low <<= 1;
				--j;
			}else{
				fibword_low <<= (j);
				break;
			}
		}	
	}
	if( len_1 >=64 ){
		bit_magic::writeIntAndMove(z, fibword_low, offset, 64);
		bit_magic::writeIntAndMove(z, fibword_high, offset, len_1 - 63);
	}else{
		bit_magic::writeIntAndMove(z, fibword_low, offset, (len_1&0x3F) +1);
	}
}

template<class int_vector1, class int_vector2>
bool fibonacci::decode(const int_vector1 &z, int_vector2 &v){
	uint64_t n = 0, carry = 0; // n = number of values to be decoded
	const uint64_t *data = z.data();
	// Determine size of v
	if( z.empty() ){// if z is empty we are ready with decoding
		v.setIntWidth(z.getIntWidth());
		v.resize(0);
		return true;
	}
	for(size_t i=0; i < (z.capacity()>>6)-1; ++i, ++data){
		n += bit_magic::b11Cnt( *data, carry );
	}
	if( z.capacity() != z.bit_size() ){
		n += bit_magic::b11Cnt( (*data) & bit_magic::Li1Mask[z.bit_size()&0x3F], carry );
	}
	else{
		n += bit_magic::b11Cnt( *data, carry );
	}
	v.setIntWidth( z.getIntWidth() ); v.resize( n );
	return decode<false, true>(z.data(), 0, n, v.begin());
}

template<bool sumup, bool increment, class Iterator>
inline uint64_t fibonacci::decode(const uint64_t *data, const size_type start_idx, size_type n, Iterator it){
	data += (start_idx >> 6);
	uint64_t w = 0, value = 0;
	int8_t buffered = 0; // bits buffered in w, in 0..64
	int8_t read = start_idx & 0x3F; // read bits in current *data 0..63
	int8_t shift = 0;
	uint32_t fibtable = 0;
	while( n ){// while not all values are decoded
		while(buffered < 13 and bit_magic::b11Cnt(w) < n ){
			w |= (((*data)>>read)<<buffered);
			if( read >= buffered ){
				++data;
				buffered += 64-read;
				read = 0;
			}else{ // read < buffered
				read += 64-buffered;
				buffered = 64;
			}
		}
		value += bit_magic::Fib2bin_0_95[(fibtable<<12) | (w&0xFFF)];
		shift  = bit_magic::Fib2binShift[w&0x1FFF];
		if( shift > 0 ){// if end of decoding
			w >>= shift;
			buffered -= shift;
			if( increment ) *(it++) = value;
			if(!sumup)
			   	value = 0;
			fibtable = 0;
			--n;
		}else{// not end of decoding
			w >>= 12;
			buffered -= 12;
			++fibtable;
		}
	}
	return value;
}

template<bool sumup, bool increment, class Iterator>
inline uint64_t fibonacci::decode1(const uint64_t *data, const size_type start_idx, size_type n, Iterator it){
	data += (start_idx >> 6);
	uint64_t w = 0, value = 0;
	int8_t buffered = 0; // bits buffered in w, in 0..64
	int8_t read = start_idx & 0x3F; // read bits in current *data 0..63
	int8_t shift = 0;
	uint32_t fibtable = 0;
	uint8_t blocknr = (start_idx>>6)%9;
	while( n ){// while not all values are decoded
		while(buffered < 13 and bit_magic::b11Cnt(w) < n ){
			w |= (((*data)>>read)<<buffered);
			if( read >= buffered ){
				++blocknr;
				++data;
				if(blocknr==8){++data;blocknr=0;}
				buffered += 64-read;
				read = 0;
			}else{ // read < buffered
				read += 64-buffered;
				buffered = 64;
			}
		}
		value += bit_magic::Fib2bin_0_95[(fibtable<<12) | (w&0xFFF)];
		shift  = bit_magic::Fib2binShift[w&0x1FFF];
		if( shift > 0 ){// if end of decoding
			w >>= shift;
			buffered -= shift;
			if( increment ) *(it++) = value;
			if(!sumup)
			   	value = 0;
			fibtable = 0;
			--n;
		}else{// not end of decoding
			w >>= 12;
			buffered -= 12;
			++fibtable;
		}
	}
	return value;
}


template<class int_vector>
bool elias_delta::encode(const int_vector &v, int_vector &z){
	z.setIntWidth( v.getIntWidth() );
//	z.m_elements = v.size();
	size_t z_bit_size = 0;
	for(typename int_vector::const_iterator it = v.begin(), end = v.end(); it != end; ++it){
		z_bit_size += encoding_length(*it);
	}
	z.bit_resize( z_bit_size ); // Initial size of z
	if( z_bit_size & 0x3F ){ // if z_bit_size % 64 != 0
		*(z.m_data + (z_bit_size>>6)) = 0; // initialize last word
	}
	z_bit_size = 0;
	uint64_t *z_data = z.m_data;
	uint8_t offset=0;
	uint64_t w;
	size_t len, len_1_len; // TODO: change to uint8_t and test it
	for(typename int_vector::const_iterator it = v.begin(), end=v.end(); it != end; ++it){
		w = *it;
		if( w == 0 ){
			throw std::logic_error("elias_delta::encode(const SDSBitVector &v, SDSBitVector &z); entry of v equals 0 that cannot be encoded!");
		}
		// (number of bits to represent w) 
		len 		= bit_magic::l1BP(w)+1;
		// (number of bits to represent the length of w) -1
		len_1_len	= bit_magic::l1BP(len);
		// Write unary representation for the length of the length of w 
		bit_magic::writeIntAndMove(z_data, 1ULL << len_1_len, offset, len_1_len+1); 
		if(len_1_len){
			bit_magic::writeIntAndMove(z_data, len, offset, len_1_len);
			bit_magic::writeIntAndMove(z_data, w, offset, len-1);
		}
	}
	return true;
}

inline void elias_delta::encode(uint64_t x, uint64_t* &z, uint8_t &offset){
	if( x == 0 ){
		throw std::logic_error("elias_delta::encode(uint64_t x, uint64_t* &z, uint8_t &offset); x equals 0 that cannot be encoded!");
	}
	uint8_t len, len_1_len;
	// (number of bits to represent w) 
	len = bit_magic::l1BP(x)+1;
	// (number of bits to represent the length of w) - 1
	len_1_len	= bit_magic::l1BP(len);
	// Write unary representation for the length of the length of w 
	bit_magic::writeIntAndMove(z, 1ULL << len_1_len, offset, len_1_len+1); 
	if(len_1_len){
		bit_magic::writeIntAndMove(z, len, offset, len_1_len);
		bit_magic::writeIntAndMove(z, x, offset, len-1);
	}
}

template<class int_vector>
bool elias_delta::decode(const int_vector &z, int_vector &v){
	size_t len_1_len, len, n = 0;
	const uint64_t *z_data	= z.data();
	const uint64_t *z_end	= z.data()+(z.bit_size()>>6);
	uint8_t offset 		= 0;
	while( (z_data < z_end) or (z_data==z_end and offset < (z.bit_size()&0x3F) )  ){
		len_1_len = bit_magic::readUnaryIntAndMove(z_data, offset);
//if(z_data==z.data() ) std::cerr<<" len_1_len="<<len_1_len<<" offset="<<(int)offset<<std::endl;		
		if( len_1_len ){
			len 	= bit_magic::readIntAndMove(z_data, offset, len_1_len) + (1 << len_1_len);
//if(z_data==z.data() ) std::cerr<<" len="<<len<<" offset="<<(int)offset<<std::endl;		
			bit_magic::move_right(z_data, offset, len-1);
		}
		++n;
	}
	v.setIntWidth( z.getIntWidth() );
//	n = z.m_elements;
	v.resize( n );
	return decode<false, true>(z.data(), 0, n, v.begin());
}

template<bool sumup, bool increment, class Iterator>
inline uint64_t elias_delta::decode(const uint64_t *data, const size_type start_idx, size_type n, Iterator it){
	data += (start_idx >> 6);
	uint64_t value = 0;
	size_type i = 0;
	size_type len_1_len, len;
	uint8_t offset = start_idx & 0x3F;
	while( i++ < n ){// while not all values are decoded
		len_1_len = bit_magic::readUnaryIntAndMove(data, offset); // read length of length of x
		if( !len_1_len ){
			value += 1;
		}else{
			len 	=  bit_magic::readIntAndMove(data, offset, len_1_len) + (1ULL << len_1_len);
			value	+= bit_magic::readIntAndMove(data, offset, len-1) + (1ULL << (len-1));
		}
		if( increment ) *(it++) = value;
		if(!sumup) value = 0;
	}
	return value;
}

template<class int_vector>
bool ternary::encode(const int_vector &v, int_vector &z){
	z.setIntWidth( v.getIntWidth() );
	size_t z_bit_size = 0;
	for(typename int_vector::const_iterator it = v.begin(), end = v.end(); it != end; ++it){
		z_bit_size += encoding_length(*it);
	}
	z.bit_resize( z_bit_size ); // Initial size of z
	if( z_bit_size & 0x3F ){ // if z_bit_size % 64 != 0
		*(z.m_data + (z_bit_size>>6)) = 0; // initialize last word
	}
	z_bit_size = 0;
	uint64_t *z_data = z.m_data, *z_temp;
	uint8_t offset=0, off_temp;
	uint64_t w=0;
	size_t len=0;
	for(typename int_vector::const_iterator it = v.begin(), end=v.end(); it != end; ++it){
		w = *it;
		if( w == 0 ){
			throw std::logic_error("ternary::encode(const SDSBitVector &v, SDSBitVector &z); entry of v equals 0 that cannot be encoded!");
		}
		// (length of bits to represent w) 
		len = encoding_length(*it)-2;
		if( (offset=(offset + len)) > 63){
			z_data += (offset>>6);
		}
		offset &= 0x3F;
		z_temp = z_data;
		off_temp = offset;
		if( (offset=(offset+2)) > 63 ){
			++z_data;
		}
		offset &= 0x3F;
//		bool not_zero = false;
		// write ternary number
		bit_magic::writeInt(z_temp, 3, off_temp, 2);
		for(int32_t i=len; i>0; i-=2){
			if( off_temp != 0 )
				off_temp -= 2;
			else{
				off_temp = 62;
				--z_temp;
			}
//			if( (w%3)>0 )
//				not_zero = true;
			bit_magic::writeInt(z_temp, w%3, off_temp, 2);
			w/=3;
		}
//		assert(not_zero);
	}
	return true;
}

template<class int_vector>
bool ternary::decode(const int_vector &z, int_vector &v){
	uint64_t n = 0; // n = number of values to be decoded
	const uint64_t *data = z.data();
	// Determine size of v
	if( z.empty() ){// if z is empty we are ready with decoding
		v.setIntWidth(z.getIntWidth());
		v.resize(0);
		return true;
	}
	for(size_t i=0; i < (z.capacity()>>6)-1; ++i, ++data){
		n += bit_magic::eB11Cnt( *data );
	}
	if( z.capacity() != z.bit_size() ){
		n += bit_magic::eB11Cnt( (*data) & bit_magic::Li1Mask[z.bit_size()&0x3F] );
	}
	else{
		n += bit_magic::eB11Cnt( *data );
	}
	v.setIntWidth( z.getIntWidth() ); v.resize( n );
	return decode<false, true>(z.data(), 0, n, v.begin());
}

template<bool sumup, bool increment, class Iterator>
inline uint64_t ternary::decode(const uint64_t *data, const size_type start_idx, size_type n, Iterator it){
	data += (start_idx >> 6);
	size_type i = 0;
	uint64_t w = 0, value = 0, tempv=0;
	uint32_t v;
	int8_t buffered = 0; // bits buffered in w, in 0..64
	int8_t read = start_idx & 0x3F; // read bits in current *data 0..63
	int8_t shift = 0;
	while( i < n ){// while not all values are decoded
		while(buffered < 16 and i + bit_magic::eB11Cnt(w) < n ){
			w |= (((*data)>>read)<<buffered);
			if( read >= buffered ){
				++data;
				buffered += 64-read;
				read = 0;
			}else{ // read < buffered
				read += 64-buffered;
				buffered = 64;
			}
		}
		v = bit_magic::Trny2bin_0_16[w&0xFFFF];
		//if shift is 16 and the most significant bit pair is not 11
		if( ((v>>13)==7) && (w&0xC000)!=0xC000 ){// not end of decoding
			tempv *= 6561; // tempv *= 3**8
			tempv += v&0x1FFF;
			w >>= 16;
			buffered -= 16;
		}else{// end of decoding
			shift = ((v>>13)+1)<<1;
			w >>= shift;
			buffered -= shift;
			tempv *= bit_magic::powerOf3[v>>13];
			tempv += v&0xFFF;
			if( sumup ) 
				value += tempv;
			else
				value = tempv;
			if( increment ) *(it++) = value;
			tempv = 0;
			++i;
		}
	}
	return value;
}



} // end namespace coder

} // end namespace sdsl

#endif
