/* sdsl - succinct data structures library
    Copyright (C) 2009 Simon Gog 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/ .
*/
/*! \file cst_sct.hpp
    \brief cst_sct.hpp contains an implementation of the compressed suffix tree (CST). 
	\author Simon Gog
*/
#ifndef INCLUDED_SDSL_CST_SCT
#define INCLUDED_SDSL_CST_SCT

#include "vectors.hpp" 
#include "algorithms.hpp"
#include "iterators.hpp"
#include "lcp_sada.hpp"
#include "balanced_parentheses_support_j.hpp"
#include "csa_sada_prac.hpp" // for std initialization of cst_sct 
#include <iostream>
#include <algorithm>
#include <cassert>
#include <cstring> // for strlen
#include <iomanip>
#include <iterator>
#include <stack> // for the calculation of the balanced parantheses sequence
#include <ostream>

namespace sdsl{

template<class Int = size_t>	
struct lcp_interval{
	Int l;
	Int left, right;
//	Int lcp_left, lcp_right;
	lcp_interval(Int l=0, Int left=0, Int right=0):l(l),left(left),right(right){};

	bool operator<(const lcp_interval &interval)const{
		if(l!=interval.l)
			return l<interval.l;
		if(left!=interval.left)
			return left<interval.left;
		return right<interval.right;

	}

	bool operator==(const lcp_interval &interval)const{
		return l==interval.l and 
			   left==interval.left and
			   right==interval.right;
	}

	bool operator!=(const lcp_interval &interval)const{
		return !(*this==interval);
	}

	lcp_interval& operator=(const lcp_interval &interval){
		l = interval.l;
		left = interval.left;
		right = interval.right;
		return *this;
	}
};

template<class Int>
inline std::ostream& operator<<(std::ostream &os, const lcp_interval<Int> &interval){
	os<<interval.l<<"-["<<interval.left<<","<<interval.right<<"]";
	return os;
}



//! A class for the Compressed Suffix Tree (CST) proposed by Ohlebusch and Gog.
/*! The CST is parameterized with an compressed suffix array, a lcp table 
 *  and a support data structure for balanced parentheses sequences.
 * 
 *  Applications of the CST: The compressed suffix tree could be used for string matching and
 *  many other application in sequence analysis. 17 applications are in the book 
 *  "Algorithms on Strings, Trees, and Sequences" of Dan Gusfield.
 */
/** \ingroup cst_group */
/* @{ */
template<class Csa = csa_sada_prac<>, class Lcp = lcp_sada<Csa>, class Bp_support = balanced_parentheses_support_j<> >
class cst_sct{
	public:
	typedef typename Csa::value_type							 value_type;	// STL Container requirement
	typedef random_access_const_iterator<cst_sct>	 			 const_iterator;// STL Container requirement
	typedef const_iterator 										 iterator;		// STL Container requirement
	typedef const value_type									 const_reference;
	typedef const_reference										 reference;
	typedef const_reference*									 pointer;
	typedef const pointer										 const_pointer;
	typedef typename Csa::size_type								 size_type;		// STL Container requirement
	typedef ptrdiff_t  											 difference_type; // STL Container requirement
	typedef Csa													 csa_type;
	typedef Bp_support											 bp_support_type;
	typedef const unsigned char*								 pattern_type;
	typedef lcp_interval<size_type>								 node_type;
	private:
		Csa 			m_csa;
		Lcp 			m_lcp;
		bit_vector		m_bp;
		Bp_support		m_bp_support;

		// TODO: auslagern in algorithm
		void construct_supercartesian_tree_bp(const Lcp &lcp, bit_vector &bp){
			bp.resize( 2*lcp.size() );    // resize bit vector for balanaced parantheses to 2 n bits
			std::stack<typename Lcp::value_type> lcp_stack;
			lcp_stack.push(0);
			bp[0] = 1; // writing an opening parenthesis
			size_type k=1;
			for(size_type i=1, l; i < lcp.size(); ++i){
				l = lcp[i];
				while( lcp_stack.size() > 0 and l < lcp_stack.top() ){
					lcp_stack.pop();
					bp[k++] = 0; // writing a closing parenthesis
				}
				lcp_stack.push(l);
				bp[k++] = 1; // writing an opening  parenthesis
			}
			while( lcp_stack.size() > 0 ){
				lcp_stack.pop();
				bp[k++] = 0; // writing a closing parenthesis
			}
			assert( k == 2*lcp.size() );
		}

		// Get the first l index of a l-[i,j] interval.
		/* I.e. given an interval [i,j], the function returns the position of the smallest entry lcp[k] with \f$ i<k\leq j \f$ 
		 * Time complexity: O(1) + O(lcp access)
		 */
		size_type get_first_l_index(size_type i, size_type j, size_type &kpos, size_type &find_close_k)const{
			if( j+1 != m_csa.size() and m_lcp[i] <= m_lcp[j+1] ){
				size_type jp1pos = m_bp_support.select(j+2);
				assert( m_bp[jp1pos] == 1 );// opening parenthesis
				assert( jp1pos > 0 and m_bp[jp1pos-1] == 0);// closing parenthesis 
				find_close_k = jp1pos-1;
				kpos 	= m_bp_support.find_open(find_close_k);
				return m_bp_support.rank( kpos )-1;
			}else{
				size_type ipos = m_bp_support.select(i+1);
				assert( m_bp[ipos] == 1 );// opening parenthesis
				ipos = m_bp_support.find_close(ipos);
				assert( m_bp[ipos] == 0 and m_bp[ipos-1] == 0 ); // closing parenthesis
				find_close_k = ipos-1;
				kpos	= m_bp_support.find_open(find_close_k);
				return m_bp_support.rank( kpos )-1;
			}	
		}

		// Get the postion of the first l index of a l-[i,j] interval in the balanced parentheses sequence.
		/* \par Time complexity
		 * 		\f$ \Order{\lcpaccess} \f$
		 */
		size_type get_pos_of_closing_parenthesis_of_the_first_l_index(size_type i, size_type j)const{
			if( j+1 != m_csa.size() and m_lcp[i] <= m_lcp[j+1] ){
				size_type jp1pos = m_bp_support.select(j+2);
				assert( m_bp[jp1pos] == 1 );// opening parenthesis
				assert( jp1pos > 0 and m_bp[jp1pos-1] == 0);// closing parenthesis 
				return jp1pos-1;
			}else{
				size_type ipos = m_bp_support.select(i+1);
				assert( m_bp[ipos] == 1 );// opening parenthesis
				ipos = m_bp_support.find_close(ipos);
				assert( m_bp[ipos] == 0 and m_bp[ipos-1] == 0 ); // closing parenthesis
				return ipos-1;
			}	
		}

		// Get the next smaller value.
		/* \par Time complexity
		 *      \f$ \Order{1} \f$
		 */
		size_type nsv(size_type i)const{
			size_type pos =  m_bp_support.select(i+1);
			size_type rank = m_bp_support.find_close(pos);
			size_type result = m_bp_support.rank(rank);
			return result;	
		}

		// Get the previous smaller value.
		/* \pre lcp[i] > 0!
		 * \par Time complexity 
		 *     \f$ \Order{\sigma \cdot \lcpaccess} \f$
		 * \sa psv
		 */  
		size_type psv_linear(size_type i, size_type lcpi)const{
//			size_type lcpi = m_lcp[i];
			assert(lcpi>0);
			size_type pos =  m_bp_support.select(i+1);
			size_type ec = m_bp_support.enclose(pos);
			size_type j = m_bp_support.rank(ec);
			while(lcpi == m_lcp[j-1]){
				ec = m_bp_support.enclose(ec);
				j = m_bp_support.rank(ec);
			}
			return j-1;
		}

		// Get the previous smaller value.
		/* \pre lcp[i] > 0!
		 * \par Time complexity 
		 *    \f$ \Order{\log\sigma \cdot \lcpaccess} \f$
		 */  
		size_type psv(size_type i, size_type lcpi)const{
			assert(lcpi!=0);// if(m_lcp[i]==0) return 0;
			size_type begin = m_bp_support.find_close(m_bp_support.select(i+1)); // calculate NSV
			size_type end = m_bp_support.rank(begin)+1;                          //     "      "
			if(end > m_csa.size() ){
				assert(end == m_csa.size()+1);
				end = m_bp.size();
			}else{
				end = m_bp_support.select(end);
			}
			if( ++begin < end){
				size_type result = m_bp_support.rank(m_bp_support.find_open(end-1))-1;
				if(m_lcp[result] < lcpi){ // TODO: explain this condition
					// binary search
					--end;
					while(begin!=end){
						size_type mid, temp;
						mid = (begin+end)>>1; // begin <= mid < end
						if( m_lcp[temp = m_bp_support.rank(m_bp_support.find_open(mid))-1] >= lcpi ){
							begin = mid+1;
						}else{
							result = temp;
							end = mid;
						}
					}
					return result;
				}
			}
			
			return m_bp_support.rank(m_bp_support.enclose(end))-1;
		}

		// Range minimum query which usese restricted enclose.
		/* \par Time complexity
		 *     \f$ \Order{\rrenclose} \f$
		 */
		size_type rmq(size_type l, size_type r)const{
			size_type i 	= m_bp_support.select(l+1);
			size_type j 	= m_bp_support.select(r+1);
			size_type fc_i 	= m_bp_support.find_close(i);
		   	if( j < fc_i ){ // i < j < find_close(j) < find_close(i) 
				return l;
			}else{ // i < find_close(i) < j < find_close(j)
				size_type ec = m_bp_support.restricted_enclose(i,j);
				if( ec == m_bp_support.size() ){// no restricted enclosing pair found
					return r;
				}else{// found restriced enclosing pair
					return m_bp_support.rank(ec)-1; // subtract 1, as the index is 0 based
				}
			}	
		}

	public:
		const Csa &csa;       // public interface to the underlying compressed suffix array
		const Lcp &lcp;       // public interface to the underlying lcp array
		const bit_vector &bp; // public interface to the underlying balanced parentheses sequence
		const Bp_support &bp_support;

		/*! \defgroup cst_sct_constructors Constructors of cst_sct */
		/* @{ */

		//! Default Constructor
		cst_sct(): csa(m_csa), lcp(m_lcp), bp(m_bp), bp_support(m_bp_support){}

		//! Constructor for the cst_sct taking a string for that the compressed suffix tree should be calculated.
		/*!
		 * \param str Text for which the \f$ \CST \f$ should be constructed. The text should be terminated by a zero byte.
		 * \pre The text has to be terminated by a zero byte.
		 * \sa cst_sct::cst_sct(const Csa &csa, const unsigned char *str) 
		 */
		cst_sct(const unsigned char *str);

		//! Constructor for the cst_sct taking an \f$\CSA\f$ and the corresponding string as argument.
		/*!
		 * \param csa  A \f$\CSA\f$  of the text str.
		 * \param str  A text that should be termindated by a zero byte.
		 * \pre The text has to be terminated by a zero byte.
		 * \sa cst_sct(const unsigned char *str) 
		 */
		cst_sct(const Csa &csa, const unsigned char *str);

		//! Copy constructor
		/*!
		 *  \param cst The cst_sct which should be copied. 
		 *  \par Time complexity
		 *       \f$ \Order{n} \f$, where \f$n=\f$cst_sct.size()
		 */
		cst_sct(const cst_sct<Csa, Lcp,Bp_support> &cst):csa(m_csa),lcp(m_lcp),bp(m_bp),bp_support(m_bp_support)
		{ 
			m_csa = cst.m_csa;
			m_lcp = cst.m_lcp;
			m_bp = cst.m_bp;
			m_bp_support = cst.m_bp_support;
			m_bp_support.set_vector(&m_bp);
		}

		/* @} */

		//! Default Destructor
		~cst_sct(){}

		//! Number of leafs of the suffix tree.
		/*! Required for the Container Concept of the STL.
		 *  \sa max_size, empty
		 */
		size_type size()const{ return m_csa.size();}

		//! Returns the largest size that cst_sct can ever have.
		/*! Required for the Container Concept of the STL.
		 *  \sa size
		 */
		static size_type max_size(){ return Csa::max_size(); }

		//! Returns if the data strucutre is empty.
		/*! Required for the Container Concept of the STL.A
		 * \sa size
		 */
		bool empty()const{ return m_csa.empty();}

		//! Swap method for cst_sct
		/*! The swap method can be defined in terms of assignment.
			This requires three assignments, each of which, for a container type, is linear
			in the container's size. In a sense, then, a.swap(b) is redundant.
			This implementation guaranties a run-time complexity that is constant rather than linear.
			\param cst cst_sct to swap.

			Required for the Assignable Conecpt of the STL.
		  */
		void swap(cst_sct<Csa, Lcp> &cst);

		//! Returns a const_iterator to the first element.
		/*! Required for the STL Container Concept.
		 *  \sa end
		 */
		const_iterator begin()const;

		//! Returns a const_iterator to the element after the last element.
		/*! Required for the STL Container Concept.
		 *  \sa begin.
		 */
		const_iterator end()const;

		//! []-operator
		/*! \param i Index of the value. \f$ i \in [0..size()-1]\f$.
		 *
		 * Required for the STL Random Access Container Concept.
		 */
		inline value_type operator[](size_type i)const;

		//! ()-operator return inverse suffix array values
		/*! \param i Index of the value. \f$ i \in [0..size()-1]\f$.
		 */
		inline value_type operator()(size_type i)const;

		//! Assignment Operator.
		/*!
		 *	Required for the Assignable Concept of the STL.
		 */
		cst_sct& operator=(const cst_sct &cst);

		//! Equality Operator
		/*! Two Instances of cst_sct are equal if
		 *  all their members are equal.
		 *  \par Required for the Equality Comparable Concept of the STL.
		 *  \sa operator!=
		 */
		bool operator==(const cst_sct &cst)const;

		//! Unequality Operator
		/*! Two Instances of cst_sct are equal if
		 *  not all their members are equal.
		 *  \par Required for the Equality Comparable Concept of the STL.
		 *  \sa operator==
		 */
		bool operator!=(const cst_sct &cst)const;

		//! Serialize to a stream.
		/*! \param out Outstream to write the data structure.
		 *  \return The number of written bytes.
		 */
		size_type serialize(std::ostream &out) const;

		//! Load from a stream.
		/*! \param in Inputstream to load the data structure from.
		 */
		void load(std::istream &in);

		/*! \defgroup cst_sct_tree_methods Tree methods of cst_sct */
		/* @{ */

		//! Return the root of the suffix tree.
		/*!
		 * \par Time complexity O(1) 
		 *      \f$ \Order{1} \f$
		 */
		node_type root() const{
			return node_type(0, 0, m_csa.size()-1 );
		}
		
		//! Decide if a node is a leaf in the suffix tree.
		/*!
		 * \param v A valid node of a cst_sct.
		 * \returns A boolean value indicating if v is a leaf. 
		 * \par Time complexity
		 *      \f$ \Order{1} \f$
		 */
		bool is_leaf(node_type v)const{
			return v.right==v.left;
		}

		//! Return the i-th leaf (1-based from left to right) of the suffix tree.
		/*!
		 * \param i 1-based position of the leaf. \f$1\leq i\leq csa.size()\f$.
		 * \return The i-th leave.
		 * \par Time complexity
		 *      \f$ \Order{1} \f$
		 * \pre \f$ 1 \leq i \leq csa.size() \f$
		 */
		node_type ith_leaf(size_t i)const{
			assert(i > 0 and i <= m_csa.size() );
			return node_type(m_csa.size()-m_csa[i-1], i-1, i-1);
		}

		//! Calculate the number of leaves in the subtree rooted at node v.
		/*! \param v A valid node of the suffix tree.
		 *  \par Time complexity
		 *       \f$ \Order{1} \f$ 
		 */
		size_type leaves_in_the_subtree(node_type v)const{
			return v.right-v.left+1;
		}

		//! Calculate the parent node of a node v.
		/*! \param v A valid node of the suffix tree. 
		 *  \return the parent node of v or the root if v==root().
		 *  \par Time complexity
		 *       \f$ \Order{\lcpaccess \cdot \log\sigma}\f$
		 */
		node_type parent(node_type v) const{
#ifndef NDEBUG
//			std::cout<<"parent interval of "<<v.l<<"-["<<v.left<<" "<<v.right<<"]"<<std::endl;
#endif		   	
			if( v.l <= 1 ){// if v.l < 1 => root() 
				return root();  // return root node
			}else{// l > 1 
				if( v.right+1 == m_csa.size() ){
					// search for the previous smaller value	
					size_type lcpi = m_lcp[v.left];
					if(lcpi==0)
						return root();
					assert(lcpi!=0);
					return node_type(lcpi, psv(v.left, lcpi), v.right);
				}else if(v.left == 0){
					// search for the next smaller value
					return node_type(m_lcp[v.right+1],0, nsv(v.right+1)-1);
				}
				else{
					size_type lcpl = m_lcp[v.left], lcpr = m_lcp[v.right+1];
					if( lcpl < lcpr ){
						return node_type(lcpr, v.left, nsv(v.right+1)-1);
					}else if(lcpl > lcpr){
						return node_type(lcpl, psv(v.left, lcpl), v.right);
					}else{
						if(lcpl==0)	
							return root();
						return node_type(lcpl, psv(v.left, lcpl), nsv(v.right+1)-1);
					}
				}
			}	
		}


		//! Get the ith child of a node v.
		/*!
		 * 	\param v A valid tree node of the cst.
		 *  \param i 1-based Index of the child which should be returned. \f$i \geq 1\f$.
		 *  \return The i-th child node of v or root() if v has no i-th child.
		 *  \par Time complexity 
		 *		  \f$ \Order{i \cdot \lcpaccess}\f$, \f$ i \leq \sigma \f$
		 *  \pre \f$ 1 \leq i \leq children(v) \f$ 
		 *  \note Could be implemented in time \f$\Order{\log i \cdot \lcpaccess}\f$ with binary search.
		 */
		node_type ith_child(node_type v, size_type i)const{
			if( is_leaf(v) ) // if v is a leave, v has no child
				return root();
			size_type k = 0, kpos = 0, k_find_close = 0;
			// v is not a leave: v has at least two children
			k = get_first_l_index(v.left, v.right, kpos, k_find_close);// get first l-index k and the position of k
			assert(m_lcp[k]==v.l);
			size_type left = v.left, right = k-1;
			size_type j=1; 
			while(j < i){
				// get next l-index m and the position of m
				size_type m    = m_bp_support.rank( m_bp_support.find_open(k_find_close-j) )-1;
				++j;
				if(m==k or v.l != m_lcp[m] ){// no next l-index exists
					if(i!=j) 
						return root();
					left 	= k;
					right 	= v.right; 
					break;
				}
				left  = k;
				right = m-1;
				k = m;
			}
			if(left==right){ // child interval is a leaf
				assert(m_csa.size()-m_csa[left] > v.l);
				return node_type(m_csa.size()-m_csa[left],left,right);			
			}else{
				// find l of the first child interval
				size_type k1 = get_first_l_index(left, right, kpos, kpos);
				assert(m_lcp[k1] > v.l);
				return node_type(m_lcp[k1], left, right);
			}
			return root();	
		}

		//! Get the number of children of a node v.
		/*! 
		 *  \param v A valid node v of a cst_sct. 
		 *  \returns The number of children of node v. 
		 *  \par Time complexity
		 *       \f$ \Order{\lcpaccess \cdot \log\sigma} \f$ 
		 */ 
		size_type children(node_type v)const{
			if( is_leaf(v) ) // if v is a leave, v has no child
				return 0;
			// v is not a leave: v has at least two children
			size_type r = get_pos_of_closing_parenthesis_of_the_first_l_index(v.left, v.right); 
			assert(m_lcp[m_bp_support.rank(m_bp_support.find_open(r))-1]==v.l);
			size_type begin = m_bp_support.preceding_closing_parentheses(r);
			assert(m_csa.sigma>0);
			if( begin > (size_type)m_csa.sigma-1 ) begin = m_csa.sigma-1;
			begin = r-begin;
			size_type end = r, mid;	
			while( end != begin ){
				mid = (begin+end)>>1;  // begin <= mid < end 
				if( m_lcp[m_bp_support.rank(m_bp_support.find_open(mid))-1] > v.l )
					begin = mid+1;
				else
					end = mid;
			}	
			return r-end+2;
		}

		//! Returns the next sibling of node v.
		/*!
		 * \param v A valid node v of the suffix tree. 
		 * \return The next (right) sibling of node v or root() if v has no next (right) sibling.
		 * \par Time complexity
		 *      \f$ \Order{\lcpaccess} \f$
		 */
		node_type sibling(node_type v)const{
			if(v == root())
				return root();
			if(v.right == m_csa.size()-1)// there exists no next sibling if the interval touches the right border
				return root();
			// now let [p..q] be the interval [v.left, v.right]
			size_type lcp_p   = m_lcp[v.left];  
			size_type lcp_qp1 = m_lcp[v.right+1]; // is defined as v.right < m_csa.size()-1
			if( lcp_p > lcp_qp1  )
				return root();
			else{ // search next l-index
				size_type new_q;
				size_type qp1_closing_pos = m_bp_support.find_close(m_bp_support.select(v.right+2));
				if( m_bp[qp1_closing_pos-1] ){// if there exists no closing parenthesis left to qp1_closing_pos
					new_q = nsv(v.right+1);   // i.e. we could not find a next l-index, only a smaller value: last sibling
					assert( new_q > v.right+1 );
				}else{
					new_q = m_bp_support.rank(m_bp_support.find_open(qp1_closing_pos-1))-1;	
					assert( new_q > v.right+1 );
					if(m_lcp[new_q] > lcp_qp1 ){ // no next l-index exists: find next smaller value, last sibling
						new_q = nsv(v.right+1);
					}else{
						assert( m_lcp[new_q] == lcp_qp1  );
					}
				}
				assert( new_q > v.right+1 );
				if( new_q-1 == v.right+1 ){ // sibling is a singlton interval
					return ith_leaf(new_q/*v.right+2*/);
				}else{
					// get l-index m of the sibling
					size_type m_pos, m_close_pos;
					size_type m = get_first_l_index(v.right+1, new_q-1, m_pos, m_close_pos);
					return node_type(m_lcp[m], v.right+1, new_q-1);
				}

			}
		}

		// Returns the next sibling of node v.
		// Only for tests.
		node_type sibling_naive(node_type v)const{
			if( v==root() )
				return root();
			node_type parent = this->parent(v);
			assert( parent != v );
			size_type nr = children(parent);
			for(size_type i=1; i <= nr; ++i)
				if( ith_child(parent, i) == v and i!=nr )
					return ith_child(parent, i+1);
			return root();
		}

		//! Get the child w of v which edge label (v,w) starts with character c.
		/*!
		 * 	\param v A valid tree node of the cst.
		 *  \param c First character of the edge label from v to the desired child.
		 *  \param char_pos Reference which will hold the position (0-based) of the matching char c in the original text.
		 *  \return The child node w which edge label (v,w) starts with c or root() if it does not exist.
		 *  \par Time complexity
		 *       \f$ \Order{ \sigma \cdot (\saaccess+\isaaccess) } \f$
		 */
		node_type child_linear(node_type v, unsigned char c, size_type &char_pos)const{
			if( is_leaf(v) ) // if v is a leave, it has no child
				return root();
			size_type k, kpos, k_find_close;
			// v is not a leave: v has at least two children
			k = get_first_l_index(v.left, v.right, kpos, k_find_close);// get first l-index k
			assert(m_lcp[k]==v.l);
			size_type left = v.left, right = k-1;
			char_pos = m_csa(m_csa[left]+v.l); // char_pos = invSA[SA[left]+v.l+1-1]
			uint16_t cc = m_csa.char2comp[c];
			size_type char_ex_max_pos = m_csa.C[cc+1], char_inc_min_pos = m_csa.C[cc];
			while( char_pos < char_ex_max_pos ){
				if( char_pos >= char_inc_min_pos ){
					goto found_child;
				}
				// get next l-index
				size_type mpos = m_bp_support.find_open((m_bp_support.find_close(kpos)-1));
				size_type m    = m_bp_support.rank( mpos )-1;
				if(m==k or v.l != m_lcp[m] ){// no next l-index exists
					left 	= k;
					right 	= v.right; 
					char_pos = m_csa(m_csa[left]+v.l);
					if( char_pos < char_inc_min_pos or char_pos >= char_ex_max_pos ) // check if character is two small
						return root();
					goto found_child;
				}
				left  = k;
				right = m-1;
				k = m;
				kpos = mpos;
				assert( m_csa(m_csa[left]+v.l) > char_pos );
				char_pos = m_csa(m_csa[left]+v.l);
			}
			return root();
found_child:			
			if(left==right){ // child interval is a leaf
				assert(m_csa.size()-m_csa[left] > v.l);
				return node_type(m_csa.size()-m_csa[left],left,right);			
			}else{
				// find l of the first child interval
				size_type k1pos, k1 = get_first_l_index(left, right, k1pos, k_find_close);
				assert(m_lcp[k1] > v.l);
				return node_type(m_lcp[k1], left, right);
			}
		}

		//! Get the child w of node v which edge label (v,w) starts with character c.
		/*!  
		 * 	\param v A valid tree node of the cst.
		 *  \param c First character of the edge label from v to the desired child.
		 *  \param char_pos Reference which will hold the position (0-based) of the matching char c in the original text.
		 *  \return The child node w which edge label (v,w) starts with c or root() if it does not exist.
		 *  \par Time complexity
		 *       \f$ \Order{(\lcpaccess+\saaccess+\isaaccess) \cdot \log\sigma} \f$ 
		 */ 
		node_type child(node_type v, const unsigned char c, size_type &char_pos)const{
			if( is_leaf(v) ) // if v is a leaf = (), v has no child
				return root();
			// else v = ( (     ))
			uint16_t cc = m_csa.char2comp[c];
			if(cc==0 and c!=0)// TODO: aendere char2comp so ab, dass man diesen sonderfall nicht braucht
				return root();
			size_type char_ex_max_pos = m_csa.C[cc+1], char_inc_min_pos = m_csa.C[cc];
			// v is not a leave: v has at least two children
			size_type k, kpos, k_find_close;
			k = get_first_l_index(v.left, v.right, kpos, k_find_close);
			assert(m_lcp[k]==v.l);
			size_type left = v.left, right = k-1;
			char_pos = m_csa(m_csa[left]+v.l);
			if( char_pos >= char_ex_max_pos ){// c is lex. greater than the first character of the first child interval
				return root();	
			}
			else if( char_pos >= char_inc_min_pos ){// found interval [left, right]
				goto found_child;
			}	
			{
				size_type k_char_pos = m_csa(m_csa[k]+v.l);
				{
					size_type begin = m_bp_support.preceding_closing_parentheses(k_find_close);
					assert(m_csa.sigma>0);
					if( begin > (size_type)m_csa.sigma-1 ) begin = m_csa.sigma-1;
					begin = k_find_close - begin;
					size_type end = k_find_close, m_find_close, m;	
					while( end != begin ){
						m_find_close = (begin+end)>>1;  // begin <= m_find_close < end 
						m = m_bp_support.rank(m_bp_support.find_open(m_find_close))-1;
						if( m_lcp[m] > v.l or /*m_lcp[m]==v.l and */ (char_pos = m_csa(m_csa[m]+v.l)) >= char_ex_max_pos ){
							begin = m_find_close+1;
						}
						else{ // m_lcp[m] == v.l and char_pos < char_ex_max_pos
							end = m_find_close;
							k = m;
							k_char_pos  = char_pos;
						}
					}	
					assert(m_lcp[k] == v.l);
					assert( m_bp_support.find_close(m_bp_support.select(k+1)) == end );
					if( k_char_pos < char_ex_max_pos and k_char_pos >= char_inc_min_pos ){// child interval exists 
						left = k;
						char_pos = k_char_pos;
						assert(end > 0);
						if( m_bp[end-1] == 1 ){ // last child interval 
							right = v.right;
						}else{
							size_type k1 = m_bp_support.rank(m_bp_support.find_open(end-1))-1;
							if( m_lcp[k1] > v.l ){// last child interval
								right = v.right;
							}else{
								right = k1-1;
							}
						}
					}
					else{
						return root();
					}
				}
			}
found_child:
			if(left==right){ // child interval is a leaf
				assert(m_csa.size()-m_csa[left] > v.l);
				return node_type(m_csa.size()-m_csa[left],left,right);			
			}else{
				// find l of the first child interval
				size_type k1pos, k1 = get_first_l_index(left, right, k1pos, k_find_close);
				assert(m_lcp[k1] > v.l);
				return node_type(m_lcp[k1], left, right);
			}
		}

	
		//! Returns the d-th character (1-based indexing) of the edge-label pointing to v.
		/*! \param v The node at which the edge path ends.
		 * \param d The position (1-based indexing) of the requested character on the edge path from the root to v. \f$ d > 0 \wedge d < v.l \f$
		 * \return The character at position d on the edge path from the root to v.
		 * \par Time complexity
		 *       \f$ \Order{ \log\sigma + (\saaccess+\isaaccess) } \f$
		 * \pre \f$ 1 \leq d \leq depth(v)  \f$ 
		 */
		unsigned char edge(node_type v, size_type d)const{
			if( d < 1 or d > depth(v) ){
				throw std::out_of_range("OUT_OF_RANGE_ERROR: cst_sct<>::edge(node_type v, size_type d). d == 0 or d > v.l!");
			}
			size_type order = m_csa(m_csa[v.left]+d-1);
			uint16_t c_begin = 1, c_end = m_csa.sigma+1, mid;
			while( c_begin < c_end ){
				mid = (c_begin+c_end)>>1;
				if( m_csa.C[mid] <= order ){
					c_begin = mid+1;
				}else{
					c_end = mid;
				}
			}
			return m_csa.comp2char[c_begin-1];
		}


		//! Calculate the lowest common ancestor (lca) of two nodes v and w of the suffix tree.
		/*!
		 * \param v The first node for which the loweset common ancestor with the second node should be computed. 
		 * \param w The second node for which the loweset common ancestor with the first node should be computed. 
		 * \return A node that is the lowest common ancestor of v and w in the suffix tree.
		 * \par Time complexity
		 *      \f$ \Order{\rrenclose}\   \f$
		 */
		node_type lca(node_type v, node_type w)const{
			if( v.left > w.left or (v.left == w.left and v.right < w.right )){
				std::swap(v, w);
			}
			assert(v.left < w.left or (v.left==w.left and v.right >= w.right));
			assert( !(v.left < w.left and v.right > w.left and v.right < w.right) ); // assert that v and w do not overlapp
			if(v.right >= w.right){// v encloses w or v==w
				return v;
			}else{// v.left < v.right < w.left < w.right
				size_type min_index = rmq(v.left+1, w.right);
/*				size_type check_min_index = v.left+1, min_value = m_lcp[v.left+1];
				for(size_type i=v.left+2;i<=w.right;++i){
					if( m_lcp[i] < min_value ){
						check_min_index = i;
						min_value = m_lcp[i];
					}
				}
				assert(min_index == check_min_index);
*/				
				size_type l = m_lcp[min_index];
				if(l==0)
					return root();
				else 
					return node_type(l, psv(min_index, l), nsv(min_index)-1);		
			}
		}

		//! Returns the string depth of node v.
		/*!
		 * \param v A valid node of a cst_sct.
		 * \returns The string depth of node v. 
		 * \par Time complexity 
		 *     \f$ \Order{1} \f$ 
		 */
		size_type depth(node_type v)const{
			return v.l;
		}

		//! Compute the suffix link of node v.
		/*!
		 * \param v A valid node of a cst_sct.
		 * \return The suffix link of node v.
		 * \par Time complexity
		 *      \f$ \Order{ \rrenclose + \lcpaccess \cdot \log\sigma} \f$
		 */
		node_type sl(node_type v)const{
			if( v == root() )
				return root();
			// get interval with first char deleted
			size_type left	 = m_csa.psi[v.left];
			if( is_leaf(v) ){
				if( v.l==1 )
					return root();
				else
					return node_type(v.l-1, left, left);
			}
			size_type right	 = m_csa.psi[v.right];
			assert(left < right);
			// rmq 
			size_type min_index = rmq(left+1, right);
			size_type l = m_lcp[min_index];
			if(l==0)
				return root();
			else
				return node_type(l, psv(min_index, l), nsv(min_index)-1);
		}

		/* @} */

};

/* @} */

// == template functions ==

template<class Csa, class Lcp, class Bp_support>
cst_sct<Csa, Lcp, Bp_support>::cst_sct(const unsigned char *str):csa(m_csa), lcp(m_lcp), bp(m_bp), bp_support(m_bp_support){
	// construct compressed suffix array 
#ifndef NDEBUG	
	std::cerr<<"create csa"<<std::endl;
#endif	
	m_csa = Csa(str);
#ifndef NDEBUG	
	std::cerr<<"csa was created"<<std::endl;
#endif	
	// construct lcp array
//	m_lcp = Lcp(&m_csa, str);
	lcp_trait<Csa, Lcp>::construct_lcp(m_lcp, m_csa, str);
#ifndef NDEBUG	
	std::cerr<<"lcp was created"<<std::endl;
#endif	

	// construct the balanced parentheses sequence  
	construct_supercartesian_tree_bp(m_lcp, m_bp);

	// construct the balanced parenthesis support for rank, select, find_close, find_open, enclose, restricted_enclose
	m_bp_support = balanced_parentheses_support_j<>(&m_bp);
}

template<class Csa, class Lcp, class Bp_support>
cst_sct<Csa, Lcp, Bp_support>::cst_sct(const Csa &f_csa, const unsigned char *str):csa(m_csa), lcp(m_lcp), bp(m_bp), bp_support(m_bp_support){
	// copy csa
	m_csa = f_csa;
#ifndef NDEBUG	
	std::cerr<<"copied csa"<<std::endl;
#endif	
	// construct lcp array
//	m_lcp = Lcp(&m_csa, str);
	lcp_trait<Csa, Lcp>::construct_lcp(m_lcp, m_csa, str);
#ifndef NDEBUG	
	std::cerr<<"calculated lcp"<<std::endl;
#endif	
	// construct the balanced parentheses sequence  
	construct_supercartesian_tree_bp(m_lcp, m_bp);
#ifndef NDEBUG	
	std::cerr<<"calculated bp"<<std::endl;
#endif	

	// construct the balanced parenthesis support for rank, select, find_close, find_open, enclose, restricted_enclose
	m_bp_support = balanced_parentheses_support_j<>(&m_bp);
#ifndef NDEBUG	
	std::cerr<<"calculated bp_support"<<std::endl;
#endif	
}



template<class Csa, class Lcp, class Bp_support>
typename cst_sct<Csa, Lcp, Bp_support>::size_type cst_sct<Csa, Lcp, Bp_support>::serialize(std::ostream &out) const{
	size_type written_bytes = 0;
	written_bytes += m_csa.serialize(out);
	written_bytes += m_lcp.serialize(out);
	written_bytes += m_bp.serialize(out);
	written_bytes += m_bp_support.serialize(out);
	return written_bytes;
}

template<class Csa, class Lcp, class Bp_support>
void cst_sct<Csa, Lcp, Bp_support>::load(std::istream &in) {
	m_csa.load(in);
	lcp_trait<Csa, Lcp>::load_lcp(m_lcp, in, m_csa);
	m_bp.load(in);
	m_bp_support.load(in, &m_bp);
}

template<class Csa, class Lcp, class Bp_support>
cst_sct<Csa, Lcp, Bp_support>& cst_sct<Csa, Lcp, Bp_support>::operator=(const cst_sct &cst){
	if(this != &cst){
		m_csa 			= cst.m_csa;
//		m_lcp 			= cst.m_lcp;
//		m_lcp.set_csa(&m_csa);
		lcp_trait<Csa, Lcp>::copy_lcp(m_lcp, cst.m_lcp, m_csa);
		m_bp  			= cst.m_bp;
	    m_bp_support 	= cst.m_bp_support;
		m_bp_support.set_vector(&m_bp);	
	}	
	return *this;
}





} // end namespace sdsl


#endif
