/* sdsl - succinct data structures library
inline rank_support_v::rank_support_v(const int_vector<1> *v, bool do_init){
    Copyright (C) 2009 Simon Gog 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/ .
*/
/* \file rank_support.hpp
    \brief rank_support.hpp contains classes that support a sdsl::bit_vector with constant time rank information. 
	\author Marco Baur, Simon Gog
*/
#ifndef INCLUDED_SDSL_RANK_SUPPORT10
#define INCLUDED_SDSL_RANK_SUPPORT10

#include "int_vector.hpp"
#include "rank_support.hpp"

//! Namespace for the succinct data structure library.
namespace sdsl{

// A class supporting rank queries for the "10" bit pattern in constant time. 
/* Space complexity: A bit vector of length n bits is supported by 0.25 n bits.
*/
class rank_support_v10: public rank_support {
	private:
	int_vector<64> m_basic_block; // basic block for interleaved storage of superblockrank and blockrank
	public:
		rank_support_v10(const int_vector<1> *v = NULL);
		rank_support_v10(const rank_support_v10 &rs);
		~rank_support_v10();

		void init(const int_vector<1>* v = 0);

		const size_type rank(size_type idx) const;

		const size_type operator()(size_type idx) const;

		size_type serialize(std::ostream &out)const;
		void load(std::istream &in, const int_vector<1> *v=NULL);
		void set_vector(const int_vector<1> *v=NULL);

		//! Assign Operator
		/*! Required for the Assignable Concept of the STL.
		 */
		rank_support_v10& operator=(const rank_support_v10 &rs);
		//! swap Operator
		/*! Swap two rank_support_v in constant time.
		 *	All members (excluded the pointer to the supported SDSBitVector) are swapped.
         *
		 *  Required for the Container Concept of the STL.
		 */
		void swap(rank_support_v10 &rs);
		//! Equality Operator
		/*! Two rank_support_vs are equal if all member variables are equal.
		 *
		 * Required for the Equality Comparable Concept of the STL.
		 * \sa operator!=
		 */
		bool operator==(const rank_support_v10 &rs)const;
		//! Unequality Operator
		/*! Two rank_support_vs are not equal if any member variable are not equal.
		 *
		 * Required for the Equality Comparable Concept of the STL.
		 * \sa operator==
		 */
		bool operator!=(const rank_support_v10 &rs)const;
};

inline rank_support_v10::rank_support_v10(const int_vector<1> *v){
	init(v);
}

inline rank_support_v10::rank_support_v10(const rank_support_v10 &rs){
	m_v = rs.m_v;
	m_basic_block = rs.m_basic_block;
}

inline void rank_support_v10::init(const int_vector<1>* v){
	set_vector(v);
	if( v == NULL or v->empty() )
		return;
	size_type basic_block_size = ((v->capacity() >> 9)+1)<<1; 
	m_basic_block.resize( basic_block_size ); // resize structure for basic_blocks
	if( m_basic_block.empty() )
		return;
	const uint64_t *data = m_v->data();
	size_type i, j=0;
	m_basic_block[0] = m_basic_block[1] = 0;
	uint64_t carry = 0;
	uint64_t sum = bit_magic::b10Cnt(*data, carry), second_level_cnt = 0;
	for(i = 1; i < (m_v->capacity()>>6) ; ++i){
		if( !(i&0x7) ){// if i%8==0
			j += 2;
			m_basic_block[j-1] = second_level_cnt;
			m_basic_block[j] 	= m_basic_block[j-2] + sum;
			second_level_cnt = sum = 0;
		}else
		{
			second_level_cnt |= sum<<(63-9*(i&0x7));//  54, 45, 36, 27, 18, 9, 0
		}
		sum += bit_magic::b10Cnt(*(++data), carry);
	}
	if( i&0x7 ){ // if i%8 != 0
		second_level_cnt |= sum << (63-9*(i&0x7));
		m_basic_block[j+1] = second_level_cnt;
	}else{ // if i%8 == 0
		j += 2;
		m_basic_block[j-1] = second_level_cnt;
	    m_basic_block[j]   = m_basic_block[j-2] + sum;
		m_basic_block[j+1] = 0;	
	}
}

inline const rank_support_v10::size_type rank_support_v10::rank(size_type idx)const{
	const uint64_t *p = m_basic_block.data() + ((idx>>8)&0xFFFFFFFFFFFFFFFEULL);// (idx/512)*2
	const uint64_t data_index= ( idx >> 6 ); // idx / 64
	// naechste Zeit kann zu Speicherverletzungen fuehren
	uint32_t carry = (data_index && (*( m_v->data() + data_index - 1 ) & (0x8000000000000000ULL))) ? 1 : 0;
	if( idx&0x3F ){ // if (idx % 64)
		return  *p + ( (*(p+1)>>(63 - 9*((idx&0x1FF)>>6) ))&0x1FF ) + 
		   		bit_magic::b1Cnt( (bit_magic:: b10Map(*(m_v->data()+data_index), carry)) & bit_magic::Li1Mask[idx&0x3F] );
	}
	else			
		return  *p + ( (*(p+1)>>(63 - 9*((idx&0x1FF)>>6) ))&0x1FF ); 
}

inline const rank_support_v10::size_type rank_support_v10::operator()(size_type idx)const{
	return rank(idx);
}


inline rank_support_v10::~rank_support_v10(){}

inline void rank_support_v10::set_vector(const int_vector<1> *v){
	m_v = v;
}

inline rank_support_v10::size_type rank_support_v10::serialize(std::ostream &out)const{
	return m_basic_block.serialize(out);
}

inline void rank_support_v10::load(std::istream &in, const int_vector<1> *v){
	set_vector(v);
	assert(m_v != NULL); // supported bit vector should be known
	m_basic_block.load(in);
}

inline rank_support_v10& rank_support_v10::operator=(const rank_support_v10 &rs){
	if(this != &rs){
		set_vector(rs.m_v);
		m_basic_block = rs.m_basic_block;
	}
	return *this;
}

inline void rank_support_v10::swap(rank_support_v10 &rs){
	if(this != &rs){ // if rs and _this_ are not the same object
		// TODO: swap m_v??? no!!!
//		std::swap(m_v, rs.m_v);
		m_basic_block.swap(rs.m_basic_block);
	}
}

inline bool rank_support_v10::operator==(const rank_support_v10 &rs)const{
	if(this == &rs)
		return true;
	return m_basic_block == rs.m_basic_block and *(rs.m_v) == *m_v;
}


inline bool rank_support_v10::operator!=(const rank_support_v10 &rs)const{
	return !(*this == rs);	
}

}// end namespace sds

#endif // end file 
